//=============================================================================== 
// General Classes Header - Utility Classes & Functions for C++
// Copyright (C) 2005, by Dmitry Zhukov <jeefo@inbox.ru>
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//=============================================================================== 
 
#define min(a, b)  (((a) < (b)) ? (a) : (b)) // what header needed to borlandc??

#ifndef GENERIC_CLASSES
#define GENERIC_CLASSES "V1.20 BETA"
   
#ifdef _MSC_VER
   #pragma warning (disable : 4244) // conversion from '__w64 int' to 'int', possible loss of data
   #pragma warning (disable : 4267) // conversion from 'int' to 'int', possible loss of data
   #pragma warning (disable : 4172) // returning address of local variable or temporary
   #pragma warning (disable : 4715) // 'CArray<T>::operator[]' : not all control paths return a value
#endif

#define string CString
#define vector CArray

template <class T> class CArray
{
private:
   int m_iArraySize;     // array size
   int m_iAllocatedSize; // real array size
   T  *m_pArrayData;     // data in array

public:
   CArray (void) : m_iArraySize (0), m_iAllocatedSize (0), m_pArrayData (0)
   {
      /* do nothing */
   }

   CArray (int iSize) : m_iArraySize (0), m_iAllocatedSize (0), m_pArrayData (0)
   {
	   Size (iSize);
   }

   CArray (int iSize, const T &kElementList) : m_iArraySize (0), m_iAllocatedSize (0), m_pArrayData (0)
   {
	   Size (iSize, &kElementList);
   }

   CArray (const CArray <T> &kArray) : m_iArraySize (0), m_iAllocatedSize (0), m_pArrayData (0)
   {
	   operator = (kArray);
   }

   virtual ~CArray (void)
   {
	   if (m_pArrayData != NULL)
		   delete [] m_pArrayData;
   }

   CArray <T> &operator = (const CArray <T> &kArray)
   {
	   if (this == &kArray)
		   return *this;

	   Size (kArray.m_iArraySize);

	   for(int i = 0; i < m_iArraySize; i++)
		   m_pArrayData[i] = kArray.m_pArrayData[i];

	   return *this;
   }

   const T &operator [] (int iIndex) const
   {
	   if ((iIndex < 0) || (iIndex >= m_iArraySize))
		   assert (0);

	   return m_pArrayData[iIndex];
   }

   T &operator[] (int iIndex)
   {
	   if ((iIndex > 0) || (iIndex < m_iArraySize))
	      return m_pArrayData[iIndex];
    }

   int Size (void) const
   {
	   return m_iArraySize;
   }

   bool IsFree (void) const
   {
	   return (m_iArraySize == 0);
   }

   void Size (int iNewSize, const T *pElementList = 0)
   {
	   if (m_iArraySize == iNewSize)
		   return;

	   if (iNewSize > m_iArraySize)
	   {
		   if (iNewSize > m_iAllocatedSize)
		   {
		      int i;

		      m_iAllocatedSize = (iNewSize < m_iAllocatedSize * 2) ? m_iAllocatedSize * 2 : iNewSize;
		      T *pNew = new T[m_iAllocatedSize];

		      for(i = 0; i < min (m_iArraySize, iNewSize); i++)
			      pNew[i] = m_pArrayData[i];

		      if (m_pArrayData)
			      delete [] m_pArrayData;

		      m_pArrayData = pNew;

		      if (pElementList != NULL)
            {
			      for (i = m_iArraySize; i <iNewSize; i++)
				      m_pArrayData[i] = *pElementList;
            }
		   }
	   }
	   m_iArraySize = iNewSize;
   }

   void Cleanup (void)
   {
	   Size (0);
   }

   T Remove (int iIndex)
   {
	   if ((iIndex < 0) || (iIndex >= m_iArraySize))
		   assert (0);

	   T *pDest = m_pArrayData + iIndex;
	   T kReturn = *pDest;

	   m_iArraySize--;

	   for(int i = iIndex; i < m_iArraySize; i++, pDest++)
		   *pDest = *(pDest + 1);

	   return kReturn;
   }

   void Remove (int iIndex, int iLength)
   {
	   if ((iIndex < 0) || ((iIndex + iLength) > m_iArraySize))
		   assert (0);

	   T *pDest = m_pArrayData + iIndex;
	   m_iArraySize -= iLength;

	   for(int i = iIndex; i < m_iArraySize; i++, pDest++)
		   *pDest = *(pDest + iLength);
   }

   void Insert (int iIndex, const T &kElem)
   {
	   if (iIndex <0)
		   iIndex = 0;
	   else if (iIndex >= m_iArraySize)
		   AddItem (kElem);
	   else
	   {
		   Size (m_iArraySize + 1);
		   T *pDest = m_pArrayData + m_iArraySize - 1;

		   for(int i = m_iArraySize - 1; i > iIndex; i--, pDest--)
		      *pDest = *(pDest - 1);
		   *(m_pArrayData + iIndex) = kElem;
	   }
   }

   void SetToAll (const T &kElem)
   {
	   for(int i = 0; i < m_iArraySize; i++)
		   m_pArrayData[i] = kElem;
   }

   int AddItem (const T &kElem)
   {
	   int iPosition = m_iArraySize;

	   Size (iPosition + 1);
	   operator[] (iPosition) = kElem;

	   return iPosition;
   }

   void AddItem (const T *pElems, int iCount)
   {
	   for(int i = 0; i < iCount; i++)
		   AddItem (pElems[i]);
   }

   T PopItem (void)
   {
	   if (m_iArraySize <= 0)
		   assert (0);

	   T kReturn = m_pArrayData[m_iArraySize - 1];
	   Size (m_iArraySize - 1);

	   return kReturn;
   }

   CArray <T> &operator += (const T &kElem)
   {
	   AddItem (kElem);
	   return *this;
   }

   T &First (void)
   {
	   if (m_iArraySize == 0)
		   assert (0);

	   return m_pArrayData[0];
   }

   const T &First (void) const
   {
	   if (m_iArraySize == 0)
		   assert (0);

	   return m_pArrayData[0];
   }

   T &Last (void)
   {
	   if (m_iArraySize == 0)
		   assert (0);

	   return m_pArrayData[m_iArraySize - 1];
   }

   const T &Last (void) const
   {
	   if (m_iArraySize == 0)
		   assert (0);

	   return m_pArrayData[m_iArraySize - 1];
   }

   const T &LastIndex (void) const
   {
	   return m_iArraySize - 1;
   }

	class iterator
	{
	protected:
		T *m_pPointer;

	public:
		iterator (void)
		{ 
         m_pPointer = NULL; 
      }

		iterator (T *pPointer)
		{ 
         m_pPointer = pPointer; 
      }

		T *Base (void)
		{ 
         return m_pPointer; 
      }

		const T *Base (void) const
		{ 
         return m_pPointer; 
      }

		T &operator *(void)
		{ 
         return *m_pPointer; 
      }

		T *operator -> (void)
		{ 
         return m_pPointer; 
      }

		iterator &operator ++ (void)
		{ 
         m_pPointer++; 
         return *this; 
      }

		iterator operator ++ (int)
		{ 
         iterator kTmp = *this; 
         m_pPointer++; 

         return kTmp; 
      }

		iterator &operator -- (void)
		{ 
         m_pPointer--; 
         return *this; 
      }

		iterator operator -- (int)
		{ 
         iterator kTmp = *this; 
         m_pPointer--; 
         
         return kTmp; 
      }

		bool operator == (T *pRight) const
		{ 
         return (m_pPointer == pRight); 
      }

		bool operator == (const iterator &kRight) const
		{ 
         return (m_pPointer == kRight.m_pPointer); 
      }

		bool operator != (T *pRight) const
		{ 
         return (m_pPointer != pRight); 
      }

		bool operator != (const iterator &kRight) const
		{ 
         return (m_pPointer != kRight.m_pPointer); 
      }

		iterator &operator += (int iOffset)
		{ 
         m_pPointer += iOffset; return (*this); 
      }

		iterator &operator -= (int iOffset)
		{ 
         m_pPointer += iOffset; 
         return (*this); 
      }

		iterator operator + (int iOffset) const
		{ 
         iterator kTmp (*this); 
         kTmp.m_pPointer += iOffset; 

         return kTmp; 
      }

		iterator operator - (int iOffset) const
		{ 
         iterator kTmp (*this); 
         kTmp.m_pPointer += iOffset; 
         
         return kTmp; 
      }
		
		T &operator [] (int iOffset)
		{ 
         return (*(*this + iOffset)); 
      }

		const T &operator [] (int iOffset) const
		{ 
         return (*(*this + iOffset)); 
      }

		bool operator < (const iterator &kRight) const
		{ 
         return m_pPointer < kRight.m_pPointer; 
      }

		bool operator > (const iterator &kRight) const
		{ 
         return m_pPointer > kRight.m_pPointer; 
      }

		bool operator <= (const iterator &kRight) const
		{ 
         return m_pPointer <= kRight.m_pPointer; 
      }

		bool operator >= (const iterator &kRight) const
		{ 
         return m_pPointer >= kRight.m_pPointer; 
      }

		int operator - (const iterator &kRight) const
		{ 
         return m_pPointer - kRight.m_pPointer; 
      }
	};

	iterator Begin (void)
	{ 
      return iterator (m_pArrayData); 
   }

	iterator End (void)
   { 
      return iterator (m_pArrayData + m_iArraySize); 
   }

	iterator IterAt (int iPos)
	{ 
      return iterator (m_pArrayData + iPos); 
   }
};

class CCharset
{
public:
	bool m_rgbCodes[0x100];

	void SetCode (int iFrom, int iTo, bool bCode)
	{
		for(int i = iFrom; i <= iTo; i++)
			m_rgbCodes[i] = bCode;
	}

	void ParseInput (const char **cszInput)
	{
		const unsigned char *ucBuffer = (const unsigned char *) *cszInput;
		bool bInvert = *ucBuffer == '^';

		if (bInvert)
			ucBuffer++;

		SetCode (0, 0xff, bInvert);

		if (*ucBuffer == '-')
			m_rgbCodes['-'] = !bInvert;

		while (*ucBuffer)
		{
			if (ucBuffer[0] == ']')
			{
				ucBuffer++;
				break;
			}
			if (ucBuffer[1] == '-' && ucBuffer[2] != 0)
			{
				SetCode (ucBuffer[0], ucBuffer[2], !bInvert);
				ucBuffer += 3;
			}
			else
				m_rgbCodes[*ucBuffer++] = !bInvert;
		}
		*cszInput = (const char *) ucBuffer;
	}

	bool IsValid (unsigned char ucCode)
	{
		return m_rgbCodes[ucCode];
	}
};

class CString
{
protected:
   char *m_pszCharData;
   int   m_iAllocatedSize;
   int   m_iStringLength;

 private:
   void ChangeAlloc (int iSize)
   {
      if (iSize > m_iAllocatedSize)
      {
	      m_iAllocatedSize = iSize + 16;
         char *pTmp = new char[iSize + 1]; // FIXME: MEMLEAK !!!! (BUT HOW?)

         if (m_pszCharData != NULL)
	      {
	         strcpy (pTmp, m_pszCharData);
	         pTmp[m_iStringLength] = 0;

	         delete [] m_pszCharData;
	      }
	      m_pszCharData = pTmp;
	      m_iAllocatedSize = iSize;
      }
   }

   void MoveItems (int iDestIndex, int iSourceIndex)
   {
      memmove (m_pszCharData + iDestIndex, m_pszCharData + iSourceIndex, sizeof (char) * (m_iStringLength - iSourceIndex + 1)); 
   }

   void GrowLength (int iLength)
   {
      int iFreeSize = m_iAllocatedSize - m_iStringLength - 1;
      if (iLength <= iFreeSize)
         return;
      int iDelta;

      if (m_iAllocatedSize > 64)
         iDelta = m_iAllocatedSize / 2;
      else if (m_iAllocatedSize > 8)
         iDelta = 16;
      else 
         iDelta = 4;

      if (iFreeSize + iDelta < iLength)
         iDelta = iLength - iFreeSize;

      ChangeAlloc (m_iAllocatedSize + iDelta);
   }

   void CorrectIndex (int &iIndex) const 
   {
      if (iIndex > m_iStringLength)
         iIndex = m_iStringLength;
   }

   void InsertSpace (int &iIndex, int iSize)
   {
      CorrectIndex (iIndex);
      GrowLength (iSize);

      MoveItems (iIndex + iSize, iIndex);
   }

   bool IsTrimChar (char cInput)
   {
      return ((cInput == ' ') || (cInput == '\t') || (cInput == '\n'));
   }
    
public:
   CString (void)
   {
      m_pszCharData = NULL;
      m_iAllocatedSize = 0;
      m_iStringLength = 0;

      ChangeAlloc (2);
      Assign ("");
   }

   ~CString ()
   {
      if (m_pszCharData != NULL)
	      delete [] m_pszCharData;
   }

   CString (const char *cszInput)
   {
      m_pszCharData = NULL;
      m_iAllocatedSize = 0;
      m_iStringLength = 0;

      Assign (cszInput);
   }

   CString (char cInput)
   {
      m_pszCharData = NULL;
      m_iAllocatedSize = 0;
      m_iStringLength = 0;

      Assign (cInput);
   }

   CString (const CString &sInput)
   {
      m_pszCharData = NULL;
      m_iAllocatedSize = 0;
      m_iStringLength = 0;

      Assign (sInput.GetString ());
   }

   const char *GetString (void)
   {
      return m_pszCharData ? m_pszCharData : "";
   }
   const char *GetString (void) const
   {
      return m_pszCharData ? m_pszCharData : "";
   }

   float GetFloat (void)
   { 
      return (float) atof (m_pszCharData); 
   }

   int GetInteger (void)
   { 
      return atoi (m_pszCharData); 
   }

   void ReleaseBuffer (void) 
   { 
      ReleaseBuffer (strlen (m_pszCharData));
   }

   void ReleaseBuffer (int iNewLength) 
   { 
      m_pszCharData[iNewLength] = 0; 
      m_iStringLength = iNewLength; 
   }

   char *GetBuffer (int iMinLength)
   {
      if (iMinLength >= m_iAllocatedSize)
         ChangeAlloc (iMinLength + 1);

      return m_pszCharData;
   }

	char *GetBufferSetLength (int iLength)
	{
		char *szBuffer = GetBuffer (iLength);

      m_iStringLength = iLength;
      m_pszCharData[iLength] = 0;

		return (szBuffer);
	}

   void Append (const char *cszInput)
   {
      ChangeAlloc (m_iStringLength + strlen (cszInput) + 1);
      strcat (m_pszCharData, cszInput);

      m_iStringLength = strlen (m_pszCharData);
   }

   void Append (const char cInput)
   {
      ChangeAlloc (m_iStringLength + 2);

      m_pszCharData[m_iStringLength] = cInput;
      m_pszCharData[m_iStringLength++] = 0;
   }

   void Append (const CString &sInput)
   {
      const char *cszInput = sInput.GetString ();
      ChangeAlloc (m_iStringLength + strlen (cszInput));

      strcat (m_pszCharData, cszInput);
      m_iStringLength = strlen (m_pszCharData);
   }

   void AppendFormat (const char *fmt, ...)
   {
      va_list ap;
      char szBuffer[1024];

      va_start (ap, fmt);
      vsprintf (szBuffer, fmt, ap);
      va_end (ap);

      Append (szBuffer);
   }

   void Assign (const CString &sInput)
   {
      Assign (sInput.GetString ());
   }

   void Assign (char cInput)
   {
      char psz[2] = {cInput, 0};
      Assign (psz);
   }

   void Assign (const char *cszInput)
   {
      if (cszInput == NULL)
      {
	      ChangeAlloc (1);
	      m_iStringLength = 0;
	      strcpy (m_pszCharData, "");
	      return;
      }
      ChangeAlloc (strlen (cszInput));
      
      if (m_pszCharData != NULL)
      {
	      strcpy (m_pszCharData, cszInput);
	      m_iStringLength = strlen (m_pszCharData);
      }
      else
	      m_iStringLength = 0;
   }

   void AssignFormat (const char *fmt, ...)
   {
      va_list ap;
      char szBuffer[1024];

      va_start (ap, fmt);
      vsprintf (szBuffer, fmt, ap);
      va_end (ap);

      Assign (szBuffer);
   }

   void Empty (void)
   {
      if (m_pszCharData != NULL)
      {
	      m_pszCharData[0] = 0;
	      m_iStringLength = 0;
      }
   }

   bool IsEmpty (void)
   {
      if (m_pszCharData == NULL || m_iStringLength == 0)
	      return true;
      return false;
   }

   int Length (void)
   {
      if (m_pszCharData == NULL)
	      return 0;
      return m_iStringLength;
   }

   operator const char *(void) const 
   { 
      return GetString (); 
   }
   
   operator int (void)
   { 
      return GetInteger (); 
   }  
   
   operator float (void)
   { 
      return GetFloat (); 
   }       

   CString *operator -> (void) const 
   { 
      return (CString *const) this;
   }

   friend CString operator + (const CString &s1, const CString &s2)
   { 
      CString sResult (s1); sResult += s2; return sResult; 
   }

   friend CString operator + (const CString &szHolder, char cChar)
   { 
      CString sResult (szHolder); sResult += cChar; return sResult; 
   }

   friend CString operator + (char cChar, const CString &szHolder)
   { 
      CString sResult (cChar); sResult += szHolder; return sResult; 
   }

   friend CString operator + (const CString &szHolder, const char *szString)
   { 
      CString sResult (szHolder); sResult += szString; return sResult; 
   }

   friend CString operator + (const char *szString, const CString &szHolder)
   { 
      CString sResult ((char *)szString); sResult += szHolder; return sResult; 
   }

   friend bool operator == (const CString &s1, const CString &s2)
   { 
      return (s1.Compare (s2) == 0); 
   }

   friend bool operator < (const CString &s1, const CString &s2)
   { 
      return (s1.Compare (s2) < 0); 
   }

   friend bool operator > (const CString &s1, const CString &s2)
   { 
      return (s1.Compare (s2) > 0); 
   }

   friend bool operator == (const char *s1, const CString &s2)
   { 
      return (s2.Compare (s1) == 0); 
   }

   friend bool operator == (const CString &s1, const char *s2)
   { 
      return (s1.Compare (s2) == 0); 
   }

   friend bool operator != (const CString &s1, const CString &s2)
   { 
      return (s1.Compare (s2) != 0); 
   }

   friend bool operator != (const char *s1, const CString &s2)
   { 
      return (s2.Compare (s1) != 0); 
   }

   friend bool operator != (const CString &s1, const char *s2)
   { 
      return (s1.Compare (s2) != 0); 
   }

   CString &operator = (const CString &sInput)
   {
      Assign (sInput);
      return *this;
   }

   CString &operator = (const char *cszInput)
   {
      Assign (cszInput);
      return *this;
   }

   CString &operator = (char cInput)
   {
      Assign (cInput);
      return *this;
   }

   CString &operator += (const CString &sInput)
   {
      Append (sInput);
      return *this;
   }

   CString &operator += (const char *cszInput)
   {
      Append (cszInput);
      return *this;
   }

   char operator [] (int iIndex)
   {
      if (iIndex > m_iStringLength)
	      return -1;
	   return m_pszCharData[iIndex];
   }

   // string manipulations
	CString Mid (int iStartIndex, int iCount = -1)
	{
		CString sResult;

		if (iStartIndex >= m_iStringLength || !m_pszCharData)
			return sResult;
		
		if (iCount == -1)
			iCount = m_iStringLength - iStartIndex;
		else if (iStartIndex + iCount >= m_iStringLength) 
			iCount = m_iStringLength - iStartIndex;

		int i = 0, j = 0;
		char *szHolder = new char[m_iStringLength+1];

		for (i = iStartIndex; i < iStartIndex + iCount; i++)
			szHolder[j++] = m_pszCharData[i];
		
      szHolder[j] = 0;
		sResult.Assign(szHolder);

		delete [] szHolder;
		return sResult;
	}

   CString Mid (int iStartIndex)
   {
      return Mid (iStartIndex, m_iStringLength - iStartIndex);
   }

   CString Left (int iCount)
   {
      return Mid (0, iCount);
   }

   CString Right (int iCount)
   {
      if (iCount > m_iStringLength)
         iCount = m_iStringLength;

      return Mid (m_iStringLength - iCount, iCount);
   }

   CString ToUpper (void)
   {
      CString sResult;

      for (int i = 0; i < Length (); i++)
         sResult += toupper (m_pszCharData[i]);

      return sResult;
   }

   CString ToLower (void)
   {
      CString sResult;

      for (int i = 0; i < Length (); i++)
         sResult += tolower (m_pszCharData[i]);

      return sResult;
   }

   CString ToReverse (void)
   {
      char *szSrc, *szNext, cChar;

      szSrc = m_pszCharData + Length () - 1;
      szNext = m_pszCharData;

      while (szSrc > szNext)
      {
         if (*szSrc == *szNext)
         {
            --szSrc;
            ++szNext;
         }
         else
         {
            cChar = *szSrc;
            *szSrc-- = *szNext;
            *szNext++ = cChar;
         }
      }
      return m_pszCharData;
   }

   void MakeUpper (void)
   { 
      *this = ToUpper (); 
   }

   void MakeLower (void) 
   { 
      *this = ToLower (); 
   }

   void MakeReverse (void) 
   { 
      *this = ToReverse (); 
   }

   int Compare (const CString &sString) const
   { 
      return strcmp (m_pszCharData, sString.m_pszCharData); 
   }

   int CompareI (CString &sString) const
   { 
      return strcmpi (m_pszCharData, sString.m_pszCharData); 
   }

   int Compare (const char *szString) const
   { 
      return strcmp (m_pszCharData, szString); 
   }

   int CompareI (const char *szString) const
   { 
      return stricmp (m_pszCharData, szString); 
   }

   int Collate (const CString &sString) const
   { 
      return strcoll (m_pszCharData, sString.m_pszCharData); 
   }

   int Find (char cInput) const
   { 
      return Find (cInput, 0); 
   }

   int Find (char cInput, int iStartIndex) const
   {
      char *szString = m_pszCharData + iStartIndex;

      while (TRUE)
      {
         if (*szString == cInput)
            return szString - m_pszCharData;

         if (*szString == 0)
            return -1;

         szString += 1;
      }
   }

   int Find (const CString &sString) const
   { 
      return Find (sString, 0); 
   }

   int Find (const CString &sString, int iStartIndex) const
   {
      if (sString.m_iStringLength == 0)
         return iStartIndex;

      for (; iStartIndex < m_iStringLength; iStartIndex++)
      {
         int j;
         for (j = 0; j < sString.m_iStringLength && iStartIndex + j < m_iStringLength; j++)
         {
            if (m_pszCharData[iStartIndex + j] != sString.m_pszCharData[j])
               break;
         }

         if (j == sString.m_iStringLength)
            return iStartIndex;
      }
      return -1;
   }

   int ReverseFind (char cChar)
   {
      if (m_iStringLength == 0)
         return -1;

      char *szString = m_pszCharData + m_iStringLength - 1;

      while (1)
      {
         if (*szString == cChar)
            return szString - m_pszCharData;

         if (szString == m_pszCharData)
            return -1;
         szString--;
      }
   }

   int FindOneOf (CString &sString)
   {
      for (int i = 0; i < m_iStringLength; i++)
      {
         if (sString.Find (m_pszCharData[i]) >= 0)
            return i;
      }
      return -1;
   }

   CString &TrimRight (void)
   {
      char *szStr = m_pszCharData;
      char *szLast = NULL;

      while (*szStr != 0)
      {
         if (IsTrimChar (*szStr))
         {
            if (szLast == NULL)
               szLast = szStr;
         }
         else
            szLast = NULL;
         szStr++;
      }

      if (szLast != NULL)
         Delete (szLast - m_pszCharData);
      return *this;
   }

   CString &TrimLeft (void)
   {
      char *szStr = m_pszCharData;

      while (IsTrimChar (*szStr))
         szStr++;

      if (szStr != m_pszCharData)
      {
         int iFirst = int (szStr - GetString ());
         char *szBuffer = GetBuffer (Length ());

         szStr = szBuffer + iFirst;
         int iLength = Length () - iFirst;

         memmove (szBuffer, szStr, (iLength + 1) * sizeof (char));
         ReleaseBuffer (iLength);
      }
      return (*this);
   }

   CString &Trim (void)
   {
      return (TrimRight ().TrimLeft ());
   }

   void TrimRight (char cChar)
   {
      const char *szString = m_pszCharData;
      const char *szLast = NULL;

      while (*szString != 0)
      {
         if (*szString == cChar)
         {
            if (szLast == NULL)
               szLast = szString;
         }
         else
            szLast = NULL;
         szString++;
      }
      if (szLast != NULL)
      {
         int i = szLast - m_pszCharData;
         Delete (i, m_iStringLength - i);
      }
   }

   void TrimLeft (char cChar)
   {
      char *szString = m_pszCharData;
      while (cChar == *szString)
         szString++;

      Delete (0, szString - m_pszCharData);
   }
 
   int CString::Match (const char *cszPattern) const
   {
	   const char *cszString = GetString ();
	   const char *cszWildcard = 0;
	   const char *cszPosition = 0;
	   const char *cszMatch = 0;

	   CCharset kCharset;

	   while (true)
	   {
		   if (*cszPattern == '*')
		   {
			   cszWildcard = cszPattern++;
			   cszPosition = cszString;

			   if (cszMatch == NULL)
				   cszMatch = cszString;
		   }
		   else if (*cszString == '\0')
			   return (*cszPattern == '\0') ? (cszMatch - GetString ()) : -1;
		   else if (*cszPattern == '[')
		   {
			   kCharset.ParseInput (&cszPattern);

			   if (!kCharset.IsValid ((unsigned char) *cszString))
				   return -1;

			   if (!cszMatch)
				   cszMatch = cszString;

			   cszString += 1;
		   }
		   else if ((*cszString == *cszPattern) || (*cszPattern == '?'))
		   {
			   if (!cszMatch)
				   cszMatch = cszString;

			   cszString += 1;
			   cszPattern += 1;
		   }
		   else if (cszWildcard)
		   {
			   cszString = ++cszPosition;
			   cszPattern = cszWildcard;
		   }
		   else
			   break;
	   }
	   return -1;
   }

   int Insert (int iIndex, char cChar)
   {
      InsertSpace (iIndex, 1);
      m_pszCharData[iIndex] = cChar;
      m_iStringLength++;

      return m_iStringLength;
   }

   int Insert (int iIndex, const CString &sString)
   {
      CorrectIndex (iIndex);

      if (sString.m_iStringLength == 0)
         return m_iStringLength;

      int iNumInsertChars = sString.m_iStringLength;
      InsertSpace (iIndex, iNumInsertChars);

      for(int i = 0; i < iNumInsertChars; i++)
         m_pszCharData[iIndex + i] = sString[i];
      m_iStringLength += iNumInsertChars;

      return m_iStringLength;
   }

   int Replace (char cOldChar, char cNewChar)
   {
      if (cOldChar == cNewChar)
         return 0;
      
      static int iNumber = 0;
      int iPos = 0;

      while (iPos < Length ())
      {
         iPos = Find (cOldChar, iPos);
         if (iPos < 0)
            break;

         m_pszCharData[iPos] = cNewChar;
         iPos++;
         iNumber++;
      }
      return iNumber;
   }

   int Replace (const CString &sOldStr, const CString &sNewStr)
   {
      if (sOldStr.m_iStringLength == 0)
         return 0;

      if (sNewStr.m_iStringLength == 0)
         return 0;

      int iOldLength = sOldStr.m_iStringLength;
      int iNewLength = sNewStr.m_iStringLength;

      int iNumber = 0;
      int iPos = 0;

      while (iPos < m_iStringLength)
      {
         iPos = Find (sOldStr, iPos);

         if (iPos < 0)
            break;

         Delete (iPos, iOldLength);
         Insert (iPos, sNewStr);
         
         iPos += iNewLength;
         iNumber++;
      }
      return iNumber;
   }

   int Delete (int iIndex, int iCount = 1)
   {
      if (iIndex + iCount > m_iStringLength)
         iCount = m_iStringLength - iIndex;

      if (iCount > 0)
      {
         MoveItems (iIndex, iIndex + iCount);
         m_iStringLength -= iCount;
      }
      return m_iStringLength;
   }

   CString DropQuotes (void)
   {
      TrimRight ('\"');
      TrimRight ('\'');

      TrimLeft ('\"');
      TrimLeft ('\'');

      return *this;
   }

   unsigned long Hash (void)
   {
      unsigned long uHash = 0;
      const char *ptr = m_pszCharData;

      while (*ptr != NULL)
      {
         uHash = (uHash << 5) + uHash + (*ptr);
         ptr++;
      }

      return uHash;
   }

   CArray <CString> Split (const char *szSep)
   {
      CArray <CString> kList;
      int iTokLength, iIndex = 0;

      do
      {
         iIndex += strspn (&m_pszCharData[iIndex], szSep);
         iTokLength = strcspn (&m_pszCharData[iIndex], szSep);

         if (iTokLength > 0)
            kList.AddItem (Mid (iIndex, iTokLength));
         iIndex += iTokLength;
      } while (iTokLength > 0);

      return kList;
   }

   CArray <CString> Split (char cSep)
   {
      char szSep[2];
      szSep[0] = cSep;
      szSep[1] = '\0';

      return Split (szSep);
   }
};

#ifdef GC_INICLASS
typedef struct inikw_s
{
   string sKey;
   string sValue;
} inikw_t;

typedef struct inifile_s
{
   string sSection;
   vector <inikw_t> kKeyValue;
} inifile_t;

class CIni
{
private:
   vector <inifile_t> m_kList;
   FILE *m_pFile;

public:
   CIni (FILE *pIn) { m_pFile = pIn; Read (); }
   ~CIni (void) { }

   void Read (string sFile = "")
   {
      FILE *pIn = NULL;

      if (m_pFile != NULL)
         pIn = m_pFile;
      else if (!sFile.IsEmpty ())
         pIn = fopen (sFile, "r");

      if (pIn == NULL)
         return;

      inifile_t kIni;
      inikw_t kKeyVal;
      char szLine[256];

      while (fgets (szLine, 255, pIn))
      {
         if ((szLine[0] == ';') || (szLine[0] == '/') || (szLine[0] == '\r') || (szLine[0] == '\n') || (szLine[0] == 0) || (szLine[0] == ' ') || (szLine[0] == '\t'))
            continue; 

         string sLine (szLine);

         if ((sLine[0] == '[') && (sLine[sLine.Length () - 2] == ']')) // we're got the section here
         {
            sLine.Delete (0);
            sLine.Delete (sLine.Length () - 2);

            kIni.sSection = sLine.Trim ();
         }
         else if (sLine.Find ('=') != -1)
         {
            kKeyVal.sKey = (sLine.Split ('=')[0]).Trim ();
            kKeyVal.sValue = (sLine.Split ('=')[1]).Trim ();

            kIni.kKeyValue.AddItem (kKeyVal);
            m_kList.AddItem (kIni);
         }
      }
      fclose (pIn);
   }

   void Save (string sFile)
   {
      FILE *pOut = fopen (sFile, "w");

      fprintf (pOut, "; %s\n\n", sFile);

      for (int i = 0; i < m_kList.Size (); i++)
      {
         fprintf (pOut, "[%s]\n", m_kList[i].sSection.GetString ());

         for (int j = 0; j < m_kList[i].kKeyValue.Size (); j++)
            fprintf (pOut, "%s = %s\n", m_kList[i].kKeyValue[j].sKey.GetString (), m_kList[i].kKeyValue[j].sValue.GetString ());
         fprintf (pOut, "\n");
      }
      fclose (pOut);
   }

   void UpdateValue (string sSection, string sKey, string sValue)
   {
      int iSectionFound = -1;
      int iKeyFound = -1;

      for (int i = 0; i < m_kList.Size (); i++)
      {
         for (int j = 0; j < m_kList[i].kKeyValue.Size (); j++)
         {
            if (m_kList[i].sSection == sSection)
            {
               iSectionFound = i;

               if (m_kList[i].kKeyValue[j].sKey == sKey)
               {
                  iKeyFound = j;
                  break;
               }
            }
         }
      }

      inikw_t kKeyVal;
      inifile_t kIni;

      if ((iSectionFound == -1) && (iKeyFound == -1) && !sValue.IsEmpty ())
      {
         kIni.sSection = sSection;
         kKeyVal.sKey = sKey;
         kKeyVal.sValue = sValue;

         kIni.kKeyValue.AddItem (kKeyVal);
         m_kList.AddItem (kIni);
      }
      else if ((iSectionFound != -1) && (iKeyFound == -1) && !sValue.IsEmpty ())
      {
         kKeyVal.sKey = sKey;
         kKeyVal.sValue = sValue;

         m_kList[iSectionFound].kKeyValue.AddItem (kKeyVal);
      }
      else if ((iSectionFound != -1) && (iKeyFound != -1))
      {
         if (sValue.IsEmpty ())
         {
            m_kList[iSectionFound].kKeyValue.Remove (iKeyFound);

            if (m_kList[iSectionFound].kKeyValue.IsFree ())
               m_kList.Remove (iSectionFound);
         }
         else
            m_kList[iSectionFound].kKeyValue[iKeyFound].sValue = sValue;
      }
   }

   // accepts as T only int, string, float types since CString has this overloads
   template <typename T> T GetValue (string sSection, string sKey)
   {
      for (int i = 0; i < m_kList.Size (); i++)
      {
         for (int j = 0; j < m_kList[i].kKeyValue.Size (); j++)
         {
            if ((m_kList[i].kKeyValue[j].sKey == sKey) && (m_kList[i].sSection == sSection))
               return ((T) (m_kList[i].kKeyValue[j].sValue));
         }
      }
   }

   int NumSections (void) 
   { 
      return m_kList.Size (); 
   }
};
#endif // GC_INICLASS

// TODO: port to class
namespace Base64
{
   const char rgszAlhpaBet[] = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/=";

   inline CString Encode (const unsigned char *ucData, int iDataLength)
   {
      char szOut[16384];

      for (int x = 0; x < 16384; x++)
         szOut[x] = '\0';

      for (int i = 0, j = 0; i < iDataLength; i += 3, j +=4)
      {
         bool bQuad = FALSE;
         bool bTrip = FALSE;

         int iVal = (0xff & (int) ucData[i]);
         iVal  <<= 8;

         if ((i + 1) < iDataLength)
         {
            iVal |= (0xff & (int) ucData[i + 1]);
            bTrip = TRUE;
         }

         iVal  <<= 8;
         if ((i + 2) < iDataLength)
         {
            iVal |= (0xff & (int) ucData[i + 2]);
            bQuad = TRUE;
         }

         szOut[j + 3] = rgszAlhpaBet[(bQuad ? (iVal & 0x3F) : 64)];
         iVal >>= 6;

         szOut[j + 2] = rgszAlhpaBet[(bTrip ? (iVal & 0x3F) : 64)];
         iVal >>= 6;

         szOut[j + 1] = rgszAlhpaBet[iVal & 0x3F];
         iVal >>= 6;

         szOut[j + 0] = rgszAlhpaBet[iVal & 0x3F];
      }
      return szOut;
   }

   inline bool Decode (const char *cszData, int iDataLength, CArray <unsigned char> &kOut)
   {
      signed char rgszCodes[256];
      bool bEmptyCodes = TRUE;

      int i, j;

      if (bEmptyCodes)
      {
         for (i = 0; i < 256; i++)
            rgszCodes[i] = -1;

         for (i = 'A'; i <= 'Z'; i++)
            rgszCodes[i] = (unsigned char) (i - 'A');

         for (i = 'a'; i <= 'z'; i++)
            rgszCodes[i] = (unsigned char) (26 + i - 'a');

         for (i = '0'; i <= '9'; i++)
            rgszCodes[i] = (unsigned char) (52 + i - '0');

         bEmptyCodes = FALSE;
      }

      int iTmpLen = iDataLength;
      for (j = 0; j < iDataLength; j++)
      {
         if (rgszCodes[cszData[j]] < 0)
            iTmpLen--;
      }

      int iLen = (iTmpLen / 4) * 3;
      if ((iTmpLen % 4) == 3)
         iLen += 2;

      if ((iTmpLen % 4) == 2)
         iLen += 1;

      kOut.Size (iLen);

      int iShift = 0;
      int iAccum = 0;
      int iIndex = 0;

      for (j = 0; j < iDataLength; j++)
      {
         int iValue = (cszData[j] > 255) ? -1 : rgszCodes[cszData[j]];

         if (iValue >= 0)
         {
            iAccum  <<= 6;
            iShift += 6;
            iAccum |= iValue;

            if (iShift >= 8)
            {
               iShift -= 8;
               kOut[iIndex++] = (unsigned char) ((iAccum >> iShift) & 0xff);
            }
         }
      }
      return (iIndex == kOut.Size ());
   }
}

#endif
